/** *******************
* Page Initialization - initialized for all supported sites upon load
******************** */

const SCRIPTS_TO_LOAD = ['js/content.js'];

let configuration;
let orangeIcon = false;
let greenIcon = false;

createContextMenu();

// Load the specified file and replace all placeholders before injecting it to the client side
const getCodeForInjection = function (filename, replacements) {
  let code = readFileFromCache(filename);
  // Replace all placeholders
  for (const key in replacements) {
    if (replacements.hasOwnProperty(key)) {
      let val = replacements[key];
      // Encode \ to \\, ' to \', trim all unnecessary white space and line breaks
      val = val.replace(/\\/g, '\\\\').replace(/'/g, "\\'").replace(/[\t\n\r]+/g, ' ');
      // Replace @@PLACEHOLDER@@ with value
      code = code.replace(`@@${key}@@`, val);
    }
  }
  return code;
};

// Inject page.js to the tab to initialize the extension
const injectPageJsToTab = function (tab) {
  // if url doesn't start with http or https I don't care ;)
  if (!tab.url.indexOf('http') == 0) {
    return;
  }
  // Load all scripts
  let code = '';
  for (const i in SCRIPTS_TO_LOAD) {
    const script = SCRIPTS_TO_LOAD[i];
    code += `${getCodeForInjection(script, {})};\n\n`;
  }

  // Inject it to page
  chrome.tabs.executeScript(tab.id, {
    code,
    runAt: 'document_start',
    allFrames: false, // not to iframes inside
  });
};

// Initialize the extension on the specified tab
const initializeTab = function (tab, startUpRetroactiveLoad) {
  // Inject page.js to the tab
  injectPageJsToTab(tab);
  setDefaultIcon();
};

// Loads all assets (JS/CSS/HTML) from disk to cache
const loadAllAssetsToCache = function (debug, callback) {
  // Load all files from disk to cache
  readFilesFromDisk(debug, SCRIPTS_TO_LOAD, () => {
    callback();
  });
};

/** *******************
* Initialization Helpers
******************** */

// Called every time a tab is updated
const onTabUpdate = function (tabId, info, tab) {
  if (info.status == 'loading') {
    initializeTab(tab, false);
  }
};

// Called every time a tab is replaced
const onTabReplace = function (addedTabId, replacedTabId) {
  chrome.tabs.get(addedTabId, (tab) => initializeTab(tab, false));
};

const onTabActivate = function (tab) {
  chrome.tabs.get(tab.tabId, (tabObject) => {
    if (tabObject != undefined && tabObject.url.indexOf('http') > -1) {
      chrome.tabs.sendMessage(tab.tabId, { mode: 'getBadge' }, (response) => {
        orangeIcon = false;
        greenIcon = false;
        if (typeof response === undefined || typeof response.badge === undefined || response.badge == '') {
          setDefaultIcon();
        } else if (response.badge == '!') {
          setOrangeIcon();
        } else if (response.badge == 'oab') {
          setBlueIcon();
        } else if (response.badge == 'ill') {
          setBlueIcon();
        } else if (response.badge == 'oc') {
          setPurpleIcon();
        } else {
          setGreenIcon();
        }
      });
    }
  });
};

// Listen to all tab events to recognize user navigation to a supported website
const addTabListeners = function () {
  if (!chrome.tabs.onUpdated.hasListener(onTabUpdate)) {
    chrome.tabs.onUpdated.addListener(onTabUpdate);
  }
  if (!chrome.tabs.onReplaced.hasListener(onTabReplace)) {
    chrome.tabs.onReplaced.addListener(onTabReplace);
  }
  if (!chrome.tabs.onActivated.hasListener(onTabActivate)) {
    chrome.tabs.onActivated.addListener(onTabActivate);
  }
};


// Retroactively initialize all existing tabs (e.g. when extension is first installed)
const retroactivelyInitExistingTabs = function () {
  chrome.windows.getAll({
    populate: true,
  }, (windows) => {
    // For all windows
    for (let i = 0; i < windows.length; i++) {
      const currentWindow = windows[i];

      // For all tabs in that window
      for (let j = 0; j < currentWindow.tabs.length; j++) {
        const currentTab = currentWindow.tabs[j];

        // Only initialize http or https pages, not ftp://, chrome:// etc.
        if (currentTab.url.match(/(http|https):\/\//gi)) {
          initializeTab(currentTab, true);
        }
      }
    }
  });
};


/** *******************
*  Initialization
******************** */

// Initialize the background script
const initializeBackgroundScript = function () {
  // Load all assets (JS/HTML/CSS) from disk to cache
  loadAllAssetsToCache(true, () => {
    // Load extension to all future tabs
    addTabListeners();
    // Load extension to all existing tabs
    retroactivelyInitExistingTabs();
  });

  setInterval(() => {
    // In dev/staging mode, reload all CSS/HTML files every second, to allow easy live development
    // without the need to reload the extension on every change
    if (ENVIRONMENT != 'production') {
      loadAllAssetsToCache(false, () => {});
    }
  }, 1000);
};

initializeBackgroundScript();

/** *******************
* set on install options page && uninstall survey
******************** */

chrome.runtime.onInstalled.addListener((object) => {
  if (object.reason != 'install') {
    return;
  }
  const optionsUrl = chrome.runtime.getURL('html/options.html');
  chrome.tabs.create({ url: optionsUrl }, (tab) => {
  // console.log('options page opened');
  });
});

chrome.runtime.setUninstallURL('https://www.oahelper.org/sorry-to-see-you-leave/');

/** *******************
* Listening to messages and take appropriate action
******************** */

chrome.runtime.onMessage.addListener((request, sender, sendResponse) => {
  let mode = '';
  switch (request.mode) {
    case 'followUrl':
      mode = 'followUrlResponse';
      followUrl(request.targetUrl, request.currentUrl).then((response) => {
        if (response) {
          setGreenIcon();
        }
        sendResponse({ mode, ona: response });
      });
      return true;
    case 'coreRecom':
      mode = 'coreRecomResponse';
      askCoreRecommendation(request.doi, request.currentUrl, request.docTitle, request.abstract).then((response) => {
        if (response.length > 0) {
          console.log(response)
          sendResponse({ mode, recommendation: response });
        } else {
          //console.log('obviously failed');
          sendResponse({ mode, recommendation: null });
        }
      });
      return true;
    case 'unpaywall':
      mode = 'unpaywallResponse';
      unpaywallCall(request.cleanedDOI).then((response) => {
        if (response.is_oa === true && response.best_oa_location != null) {
          response.best_oa_location.source = 'Unpaywall.org';
          response.best_oa_location.title = 'Open Access Version found from unpaywall.org';
          setOrangeIcon();
        } else {
          setDefaultIcon();
        }
        sendResponse({ mode, data: response });
      });
      return true;
    case 'coreoa':
      mode = 'coreoaResponse';
      coreOpenAccessCall(request.cleanedDOI).then((response) => {
        sendResponse({ mode, data: response });
      });
      return true;
    case 'citationcount':
      mode = 'citationcountResponse';
      openCitationsCounts(request.cleanedDOI).then((response) => {
        let citationCount = 0;
        if (response[0] !== undefined && response[0].count !== undefined && response[0].count > 0) {
          citationCount = response[0].count;
          setPurpleIcon();
        }
        sendResponse({ mode, citationCount });
      });
      return true;
    case 'getoptions':
      sendResponse({ mode: 'options', options: configuration });
      return true;
    case 'setBlueIcon':
      setBlueIcon();
      sendResponse({ mode: 'blueSet' });
    case 'getProxyDomains':
      if(request.id != undefined && request.domainurl != undefined) {
        doGetProxyDomains(request.id, request.domainurl);
        sendResponse({ mode: 'proxydomainsdone'});
      }
    case 'openTab':
      if(request.url != undefined && request.url != '') {
        chrome.tabs.create({ url: request.url });
      }
      
    default:
      return true;
  }
});


/** *******************
* Monitor storage and reload the extension
******************** */

chrome.storage.onChanged.addListener((changes, namespace) => {
  chrome.storage.local.get(['core', 'core1', 'oab', 'ezproxy', 'url', 'search', 'opencitation', 'nobadge', 'ill', 'illurl', 'name', 'id', 'consolelog', 'domainurl', 'lastload', 'supportedDomains'], (items) => {
    configuration = items;
    updateContextMenu();
  });
});

/** *******************
* Context Menu Support
******************** */

const CONTEXT_MENU_ID = 'oahelper_context';


function getword(info, tab) {
  if (info.menuItemId !== CONTEXT_MENU_ID) {
    return;
  }
  chrome.storage.local.get(['core', 'core1', 'oab', 'ezproxy', 'url', 'search'], (items) => {
    if (items.search == 'google') {
      chrome.tabs.create({
        url: `https://scholar.google.com/scholar?q=${info.selectionText}`,
      });
    } else if (items.search == 'microsoft') {
      chrome.tabs.create({
        url: `https://academic.microsoft.com/search?q=${info.selectionText}`,
      });
    } else if (items.search == 'core') {
      chrome.tabs.create({
        url: `https://core.ac.uk/search?q=${info.selectionText}`,
      });
    }
    else if (items.search == 'base') {
      chrome.tabs.create({
        url: `https://www.base-search.net/Search/Results?lookfor=${info.selectionText}&name=&oaboost=1&newsearch=1&refid=dcbasen`,
      });
    }
    else if (items.search == 'dimensions') {
      chrome.tabs.create({
        url: `https://app.dimensions.ai/discover/publication?search_mode=content&search_text=${info.selectionText}&search_type=kws&search_field=full_search`,
      });
    }
    else if (items.search == 'semantic') {
      chrome.tabs.create({
        url: `https://www.semanticscholar.org/search?q=${info.selectionText}&sort=relevance`,
      });
    }
    
  });
}

function updateContextMenu() {
  chrome.contextMenus.onClicked.removeListener(getword);
  if (configuration.search == 'google') {
    chrome.contextMenus.update(CONTEXT_MENU_ID, {
      title: 'Search Google Scholar: "%s"',
      contexts: ['selection'],
    });
  } else if (configuration.search == 'microsoft') {
    chrome.contextMenus.update(CONTEXT_MENU_ID, {
      title: 'Search MS Academic: "%s"',
      contexts: ['selection'],
    });
  } else if (configuration.search == 'core') {
    chrome.contextMenus.update(CONTEXT_MENU_ID, {
      title: 'Search core.ac.uk: "%s"',
      contexts: ['selection'],
    });
  }else if (configuration.search == 'base') {
    chrome.contextMenus.update(CONTEXT_MENU_ID, {
      title: 'Search base-search.net: "%s"',
      contexts: ['selection'],
    });
  }else if (configuration.search == 'dimensions') {
    chrome.contextMenus.update(CONTEXT_MENU_ID, {
      title: 'Search dimensions.ai: "%s"',
      contexts: ['selection'],
    });
  }else if (configuration.search == 'semantic') {
    chrome.contextMenus.update(CONTEXT_MENU_ID, {
      title: 'Search samanticscholar.org: "%s"',
      contexts: ['selection'],
    });
  }
  chrome.contextMenus.onClicked.addListener(getword);
}

function createContextMenu() {
  chrome.storage.local.get(['core', 'core1', 'oab', 'ezproxy', 'url', 'search', 'opencitation', 'nobadge', 'ezproxy', 'url', 'ill', 'illurl', 'name', 'id', 'consolelog', 'domainurl', 'lastload', 'supportedDomains'], (items) => {
    configuration = items;
    if (items.domainurl != undefined) {
      doSync(items.id, items.lastload);
    }
    if (items.search == 'google') {
      chrome.contextMenus.create({
        title: 'Search Google Scholar: "%s"',
        contexts: ['selection'],
        id: CONTEXT_MENU_ID,
      });
    } else if (items.search == 'microsoft') {
      chrome.contextMenus.create({
        title: 'Search MS Academic: "%s"',
        contexts: ['selection'],
        id: CONTEXT_MENU_ID,
      });
    } else if (items.search == 'core') {
      chrome.contextMenus.create({
        title: 'Search core.ac.uk: "%s"',
        contexts: ['selection'],
        id: CONTEXT_MENU_ID,
      });
    }else if (items.search == 'base') {
      chrome.contextMenus.create({
        title: 'Search base-search.net: "%s"',
        contexts: ['selection'],
        id: CONTEXT_MENU_ID,
      });
    }else if (items.search == 'dimensions') {
      chrome.contextMenus.create({
        title: 'Search dimensions.ai: "%s"',
        contexts: ['selection'],
        id: CONTEXT_MENU_ID,
      });
    }else if (items.search == 'semantic') {
      chrome.contextMenus.create({
        title: 'Search semanticscholar.org: "%s"',
        contexts: ['selection'],
        id: CONTEXT_MENU_ID,
      });
    }
    else{
      chrome.contextMenus.create({
      title: 'Search Google Scholar: "%s"',
      contexts: ['selection'],
      id: CONTEXT_MENU_ID,
    });
  }
    chrome.contextMenus.onClicked.addListener(getword);
  });
}

/** *******************
* fundamental extension functionality
******************** */

async function followUrl(targetUrl, currentUrl) {
  const turl = new URL(targetUrl);
  const domain0 = turl.host;
  const cUrl1 = new URL(currentUrl);
  const domain1 = cUrl1.host;

  if (domain0 === domain1) {
    return true;
  }

  const response = await fetch(targetUrl);
  const finalUrlString = response.url;
  let myFinalUrl = finalUrlString.replace('https://', '');
  myFinalUrl = myFinalUrl.replace('http://', '');
  let myCurrent = currentUrl.replace('https://', '');
  myCurrent = myCurrent.replace('http://', '');


  const next1 = new URL(finalUrlString);

  const domain2 = next1.host;

  if (myFinalUrl === myCurrent) {
    return true;
  }
  if (domain1 === domain2) {
    return true;
  }
  if (domain1 === 'www.sciencedirect.com' && domain2 === 'linkinghub.elsevier.com') {
    return true;
  }
  if (domain1 === 'psycnet.apa.org' && domain2 === 'doi.apa.org') {
    return true;
  }
  if (currentUrl.indexOf('www.ncbi.nlm.nih.gov/pmc/') > -1) {
    return true;
  }

  return false;
}

async function askCoreRecommendation(mydoi, currentUrl, docTitle, abstract) {
  const coreApiKey = 'uMPbcSzO7HNK8wDF5h2gAst3xjrldXEo';
  const url = 'https://api.core.ac.uk/v3/recommend';
  const jsonPayload = `{
    "limit": 3,
    "identifier": "${mydoi}",
    "abstract": "${abstract}",
    "authors": "",
    "title": "${docTitle}",
    "exclude": [ "fullText"]
    }`;

  try {
    const response = await fetch(url, {
      method: 'POST',
      mode: 'cors',
      cache: 'no-cache',
      credentials: 'same-origin',
      headers: {
        'Authorization': `Bearer ${coreApiKey}`,
      },
      redirect: 'follow',
      referrerPolicy: 'no-referrer',
      body: jsonPayload,
    });
    return response.json();
  } catch (err) {
    return { msg: 'failed', data: [] };
  }
}

async function unpaywallCall(cleanedDOI) {
  const url = `https://api.unpaywall.org/v2/${cleanedDOI}?email=info@oahelper.org`;
  try {
    const response = await fetch(url);
    if(response.ok){
      return response.json();
    }
    else{
      return [];
    }
  } catch (err) {
    return [];
  }
}


async function coreOpenAccessCall(cleanedDOI) {
  const coreApiKey = 'uMPbcSzO7HNK8wDF5h2gAst3xjrldXEo';
  const url = `https://api.core.ac.uk/v3/discover`;
  const jsonPayload = `{"doi" : "${cleanedDOI}"}`;
  try {
    const response = await fetch(url, {
      method: 'POST',
      mode: 'cors',
      cache: 'no-cache',
      credentials: 'same-origin',
      headers: {
        'Authorization': `Bearer ${coreApiKey}`,
      },
      redirect: 'follow',
      referrerPolicy: 'no-referrer',
      body: jsonPayload,
    });
    if(response.ok){
      return response.json();
    }
    else{
      return [];
    }
  } catch (err) {
    return [];
  }
}

async function openCitationsCounts(cleanedDOI) {
  const url = `https://opencitations.net/index/api/v1/citation-count/${cleanedDOI}`;
  try {
    const response = await fetch(url);
    if(response.ok){
      return response.json();
    }
    else{
      return [];
    }
  } catch (err) {
    return [];
  }
}

function doSync(id, lastload){
  //console.log(`doSync called ${Date.now()}`);
  // get timestamp 7 days ago
  let checkTimeStamp = Math.floor(Date.now() / 1000) - (7*24*60*60);
  // if last load is greater 7 days ago, we will reload settings
  if (lastload < checkTimeStamp) {
    reloadSettings(id)
  }
  else{
    console.log('doSync doesn\'t need to run');
  }
}

function reloadSettings(id) {
    doProxySearchById(id).then((response) => {
      if (response != undefined && (response.code >= 200 && response.code <= 299)) {
        if (response.data.length == 0) {
          // we found nothing, need to give up
          //console.log('background / reloadSettings found no settings')
          return;
        } else{
          //console.log('background / reloadSettings has found settings and will attempt to save next');
          // we found one, as we search by id there can only be one result
          saveInstitute(response.data[0].id, response.data[0].institution, response.data[0].proxyUrl.replace('{targetUrl}', ''), response.data[0].ill.replace('{doi}', ''), response.data[0].domainUrl);
        } 
      } else {
        //console.log('background / doGetProxyDomains had an error in the doProxySearch Response, response undefined, or response.code outside valid range');
        // there was an error, we need to give up
        return;
      }
    });
}

async function doProxySearchById(searchTerm) {
  const url = 'https://www.oahelper.org/backend/institutes/api.php';
  const apiKey = 'l6sNUJ3Knh';
  const formData = new URLSearchParams();
  let searchtype = 'id';
  formData.append(searchtype, searchTerm);
  const response = await fetch(url, {
    method: 'POST',
    mode: 'cors',
    cache: 'no-cache',
    credentials: 'same-origin',
    headers: {
	  'Content-Type': 'application/x-www-form-urlencoded',
	  'X-Token': apiKey,
    },
    redirect: 'follow',
    referrerPolicy: 'no-referrer',
    body: formData,
  });
  return response.json();
}

function saveInstitute(id, name, url, illurl, domainurl) {
  let lastload = Math.floor(Date.now() / 1000);
  let ill = false;
  if(illurl.length > 0 && illurl.indexOf('http') === 0) {
    ill = true;
  }
  let ezproxy = false;
  if(url.length > 0 && url.indexOf('http') === 0){
    ezproxy = true;
  }
  chrome.storage.local.set({
    id,
    name,
    url,
    illurl,
    ezproxy,
    ill,
    domainurl,
    lastload
  }, () => {
    //console.log('background / saveInstitute has finished saving');
    // we just reloaded the settings, now let's check if we should obtain proxy domains and then run that process
    if (domainurl != '') {
      doGetProxyDomains(id, domainurl)
    }
  });
}

function doGetProxyDomains(id, domainurl){
  if (domainurl == '' || domainurl == "undefined" || domainurl == undefined){
    return;
  }
  getProxyDomains(domainurl).then((response) => {
    if (response.legnth != 0 && id == response.id && response.domains.length > 0) {
      // we need to save the domains now
      chrome.storage.local.set({
        supportedDomains: response.domains,
        lastLoad: Math.floor(Date.now() / 1000),
        ezproxy: true,
        url: response.prefix
      })
      // we might want to log something for now
      //console.log('background / doGetProxyDomains has finished');
    }
  })
}

async function getProxyDomains(domainurl) {
  try {
    const response = await fetch(domainurl);
    if(response.ok){
      return response.json();
    }
    else{
      return [];
    }
  } catch (err) {
    return [];
  }
}

function setGreenIcon() {
  greenIcon = true;
  chrome.browserAction.setIcon({
    path: {
      19: 'images/oahelper_green_19.png',
      38: 'images/oahelper_green_38.png',
    },
  });
}

function setOrangeIcon() {
  orangeIcon = true;
  chrome.browserAction.setIcon({
    path: {
      19: 'images/oahelper_orange_19.png',
      38: 'images/oahelper_orange_38.png',
    },
  });
}

function setBlueIcon() {
  if(orangeIcon || greenIcon){
    return;
  }
  chrome.browserAction.setIcon({
    path: {
      19: 'images/oahelper_blue_19.png',
      38: 'images/oahelper_blue_38.png',
    },
  });
}

function setPurpleIcon() {
  if(orangeIcon || greenIcon){
    return;
  }
  chrome.browserAction.setIcon({
    path: {
      19: 'images/oahelper_purple_19.png',
      38: 'images/oahelper_purple_38.png',
    },
  });
}

function setDefaultIcon() {
  chrome.browserAction.setIcon({
    path: {
      19: 'images/oahelper_black_19.png',
      38: 'images/oahelper_black_38.png',
    },
  });
}